# Form implementation generated from reading ui file 'stashdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_StashDialog(object):
    def setupUi(self, StashDialog):
        StashDialog.setObjectName("StashDialog")
        StashDialog.setWindowModality(Qt.WindowModality.NonModal)
        StashDialog.setEnabled(True)
        StashDialog.resize(461, 432)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(StashDialog.sizePolicy().hasHeightForWidth())
        StashDialog.setSizePolicy(sizePolicy)
        StashDialog.setSizeGripEnabled(False)
        StashDialog.setModal(True)
        self.vboxlayout = QVBoxLayout(StashDialog)
        self.vboxlayout.setObjectName("vboxlayout")
        self.messageEdit = QLineEdit(parent=StashDialog)
        self.messageEdit.setObjectName("messageEdit")
        self.vboxlayout.addWidget(self.messageEdit)
        self.keepCheckBox = QCheckBox(parent=StashDialog)
        self.keepCheckBox.setChecked(False)
        self.keepCheckBox.setObjectName("keepCheckBox")
        self.vboxlayout.addWidget(self.keepCheckBox)
        spacerItem = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.vboxlayout.addItem(spacerItem)
        self.willStashLabel = QLabel(parent=StashDialog)
        self.willStashLabel.setText("(&WILL STASH THESE:)")
        self.willStashLabel.setWordWrap(True)
        self.willStashLabel.setObjectName("willStashLabel")
        self.vboxlayout.addWidget(self.willStashLabel)
        self.fileList = QListWidget(parent=StashDialog)
        self.fileList.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.fileList.setIconSize(QSize(16, 16))
        self.fileList.setObjectName("fileList")
        self.vboxlayout.addWidget(self.fileList)
        self.indexAndWtWarning = QLabel(parent=StashDialog)
        self.indexAndWtWarning.setWordWrap(True)
        self.indexAndWtWarning.setObjectName("indexAndWtWarning")
        self.vboxlayout.addWidget(self.indexAndWtWarning)
        self.buttonBox = QDialogButtonBox(parent=StashDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)
        self.willStashLabel.setBuddy(self.fileList)

        self.retranslateUi(StashDialog)
        self.buttonBox.rejected.connect(StashDialog.reject)
        self.buttonBox.accepted.connect(StashDialog.accept)

    def retranslateUi(self, StashDialog):
        StashDialog.setWindowTitle(_p("StashDialog", "New stash"))
        self.messageEdit.setPlaceholderText(_p("StashDialog", "Optional stash message"))
        self.keepCheckBox.setToolTip(_p("StashDialog", "Normally, stashed changes are removed from the working directory. Tick this to leave the stashed changes intact instead."))
        self.keepCheckBox.setText(_p("StashDialog", "&Retain stashed changes in working directory"))
        self.indexAndWtWarning.setText(_p("StashDialog", "Warning: Some of the files have both staged and unstaged changes. Those changes will be combined in the stash."))
