# Form implementation generated from reading ui file 'searchbar.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_SearchBar(object):
    def setupUi(self, SearchBar):
        SearchBar.setObjectName("SearchBar")
        SearchBar.resize(329, 26)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SearchBar.sizePolicy().hasHeightForWidth())
        SearchBar.setSizePolicy(sizePolicy)
        SearchBar.setAutoFillBackground(True)
        self.horizontalLayout = QHBoxLayout(SearchBar)
        self.horizontalLayout.setContentsMargins(2, 2, 2, 2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.providerChooser = QToolButton(parent=SearchBar)
        self.providerChooser.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.providerChooser.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.providerChooser.setObjectName("providerChooser")
        self.horizontalLayout.addWidget(self.providerChooser)
        self.lineEdit = QLineEdit(parent=SearchBar)
        self.lineEdit.setClearButtonEnabled(True)
        self.lineEdit.setObjectName("lineEdit")
        self.horizontalLayout.addWidget(self.lineEdit)
        self.filterCheckBox = QCheckBox(parent=SearchBar)
        self.filterCheckBox.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.filterCheckBox.setObjectName("filterCheckBox")
        self.horizontalLayout.addWidget(self.filterCheckBox)
        self.forwardButton = QToolButton(parent=SearchBar)
        self.forwardButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.forwardButton.setText("↓")
        self.forwardButton.setObjectName("forwardButton")
        self.horizontalLayout.addWidget(self.forwardButton)
        self.backwardButton = QToolButton(parent=SearchBar)
        self.backwardButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.backwardButton.setText("↑")
        self.backwardButton.setObjectName("backwardButton")
        self.horizontalLayout.addWidget(self.backwardButton)
        self.closeButton = QToolButton(parent=SearchBar)
        self.closeButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.closeButton.setObjectName("closeButton")
        self.horizontalLayout.addWidget(self.closeButton)

        self.retranslateUi(SearchBar)

    def retranslateUi(self, SearchBar):
        self.providerChooser.setToolTip(_p("SearchBar", "Search scope"))
        self.filterCheckBox.setToolTip(_p("SearchBar", "Show only matching rows"))
        self.filterCheckBox.setText(_p("SearchBar", "Filter"))
        self.forwardButton.setToolTip(_p("SearchBar", "Next Occurrence"))
        self.backwardButton.setToolTip(_p("SearchBar", "Previous Occurrence"))
        self.closeButton.setToolTip(_p("SearchBar", "Close Search Bar"))
        self.closeButton.setText(_p("SearchBar", "Done"))
