# Form implementation generated from reading ui file 'ignorepatterndialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_IgnorePatternDialog(object):
    def setupUi(self, IgnorePatternDialog):
        IgnorePatternDialog.setObjectName("IgnorePatternDialog")
        IgnorePatternDialog.resize(514, 131)
        self.formLayout = QFormLayout(IgnorePatternDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QLabel(parent=IgnorePatternDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.label)
        self.label_2 = QLabel(parent=IgnorePatternDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.label_2)
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.formLayout.setItem(2, QFormLayout.ItemRole.FieldRole, spacerItem)
        self.buttonBox = QDialogButtonBox(parent=IgnorePatternDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.fileEdit = QComboBoxWithPreview(parent=IgnorePatternDialog)
        self.fileEdit.setObjectName("fileEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.fileEdit)
        self.patternEdit = QComboBoxWithPreview(parent=IgnorePatternDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.patternEdit.sizePolicy().hasHeightForWidth())
        self.patternEdit.setSizePolicy(sizePolicy)
        self.patternEdit.setEditable(True)
        self.patternEdit.setObjectName("patternEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.patternEdit)
        self.label.setBuddy(self.patternEdit)
        self.label_2.setBuddy(self.fileEdit)

        self.retranslateUi(IgnorePatternDialog)
        self.buttonBox.accepted.connect(IgnorePatternDialog.accept)
        self.buttonBox.rejected.connect(IgnorePatternDialog.reject)
        IgnorePatternDialog.setTabOrder(self.patternEdit, self.fileEdit)

    def retranslateUi(self, IgnorePatternDialog):
        IgnorePatternDialog.setWindowTitle(_p("IgnorePatternDialog", "Ignore file name pattern"))
        self.label.setText(_p("IgnorePatternDialog", "&Pattern:"))
        self.label_2.setText(_p("IgnorePatternDialog", "&Add to:"))
from gitfourchette.toolbox.qcomboboxwithpreview import QComboBoxWithPreview
