# Form implementation generated from reading ui file 'repostub.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_RepoStub(object):
    def setupUi(self, RepoStub):
        RepoStub.setObjectName("RepoStub")
        RepoStub.resize(670, 327)
        RepoStub.setWindowTitle("RepoStub")
        self.verticalLayout = QVBoxLayout(RepoStub)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.stackedWidget = QStackedWidget(parent=RepoStub)
        self.stackedWidget.setObjectName("stackedWidget")
        self.promptPage = QWidget()
        self.promptPage.setObjectName("promptPage")
        self.gridLayout = QGridLayout(self.promptPage)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QSpacerItem(20, 132, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem, 0, 1, 1, 1)
        self.promptIcon = QLabel(parent=self.promptPage)
        self.promptIcon.setText("ICON")
        self.promptIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptIcon.setObjectName("promptIcon")
        self.gridLayout.addWidget(self.promptIcon, 1, 1, 1, 1)
        self.promptNameLabel = QLabel(parent=self.promptPage)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.promptNameLabel.sizePolicy().hasHeightForWidth())
        self.promptNameLabel.setSizePolicy(sizePolicy)
        self.promptNameLabel.setMaximumSize(QSize(300, 16777215))
        self.promptNameLabel.setText("Long Long Long Long Long Long Long Long Long Long Name")
        self.promptNameLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptNameLabel.setWordWrap(True)
        self.promptNameLabel.setObjectName("promptNameLabel")
        self.gridLayout.addWidget(self.promptNameLabel, 2, 1, 1, 1)
        self.promptReadyLabel = QLabel(parent=self.promptPage)
        self.promptReadyLabel.setMinimumSize(QSize(300, 0))
        self.promptReadyLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.promptReadyLabel.setWordWrap(True)
        self.promptReadyLabel.setObjectName("promptReadyLabel")
        self.gridLayout.addWidget(self.promptReadyLabel, 3, 1, 1, 1)
        spacerItem1 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem1, 4, 0, 1, 1)
        self.promptLoadButton = QPushButton(parent=self.promptPage)
        self.promptLoadButton.setObjectName("promptLoadButton")
        self.gridLayout.addWidget(self.promptLoadButton, 4, 1, 1, 1)
        spacerItem2 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem2, 4, 2, 1, 1)
        spacerItem3 = QSpacerItem(20, 132, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem3, 5, 1, 1, 1)
        self.stackedWidget.addWidget(self.promptPage)
        self.progressPage = QWidget()
        self.progressPage.setObjectName("progressPage")
        self.gridLayout_2 = QGridLayout(self.progressPage)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem4 = QSpacerItem(20, 154, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout_2.addItem(spacerItem4, 0, 1, 1, 1)
        self.progressLabel = QLabel(parent=self.progressPage)
        self.progressLabel.setMinimumSize(QSize(300, 0))
        self.progressLabel.setText("PROGRESS MESSAGE Long Long Long Long Long Long Long Long Long Long Name")
        self.progressLabel.setWordWrap(True)
        self.progressLabel.setObjectName("progressLabel")
        self.gridLayout_2.addWidget(self.progressLabel, 1, 1, 1, 2)
        spacerItem5 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem5, 2, 0, 1, 1)
        self.progressBar = QProgressBar(parent=self.progressPage)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout_2.addWidget(self.progressBar, 2, 1, 1, 1)
        self.progressAbortButton = QPushButton(parent=self.progressPage)
        self.progressAbortButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.progressAbortButton.setText("")
        self.progressAbortButton.setObjectName("progressAbortButton")
        self.gridLayout_2.addWidget(self.progressAbortButton, 2, 2, 1, 1)
        spacerItem6 = QSpacerItem(190, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem6, 2, 3, 1, 1)
        spacerItem7 = QSpacerItem(20, 154, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout_2.addItem(spacerItem7, 3, 1, 1, 1)
        self.stackedWidget.addWidget(self.progressPage)
        self.verticalLayout.addWidget(self.stackedWidget)

        self.retranslateUi(RepoStub)
        self.stackedWidget.setCurrentIndex(0)

    def retranslateUi(self, RepoStub):
        self.promptReadyLabel.setText(_p("RepoStub", "Ready to load this repository."))
        self.promptLoadButton.setText(_p("RepoStub", "Load"))
        self.progressAbortButton.setToolTip(_p("RepoStub", "Abort"))
