# Form implementation generated from reading ui file 'donateprompt.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_DonatePrompt(object):
    def setupUi(self, DonatePrompt):
        DonatePrompt.setObjectName("DonatePrompt")
        DonatePrompt.resize(475, 360)
        DonatePrompt.setWindowTitle("")
        self.verticalLayout_2 = QVBoxLayout(DonatePrompt)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(20)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.label = QLabel(parent=DonatePrompt)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.mugshot = QLabel(parent=DonatePrompt)
        self.mugshot.setText("MUG")
        self.mugshot.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.mugshot.setObjectName("mugshot")
        self.horizontalLayout.addWidget(self.mugshot)
        spacerItem2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        spacerItem3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem3)
        self.stackedWidget = QStackedWidget(parent=DonatePrompt)
        self.stackedWidget.setObjectName("stackedWidget")
        self.page = QWidget()
        self.page.setObjectName("page")
        self.gridLayout_2 = QGridLayout(self.page)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem4, 3, 4, 1, 1)
        self.donateButton = QPushButton(parent=self.page)
        self.donateButton.setDefault(True)
        self.donateButton.setObjectName("donateButton")
        self.gridLayout_2.addWidget(self.donateButton, 1, 1, 1, 3)
        self.byeButton = QPushButton(parent=self.page)
        self.byeButton.setFlat(True)
        self.byeButton.setObjectName("byeButton")
        self.gridLayout_2.addWidget(self.byeButton, 3, 2, 1, 1)
        spacerItem5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem5, 3, 1, 1, 1)
        spacerItem6 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem6, 3, 3, 1, 1)
        spacerItem7 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem7, 3, 0, 1, 1)
        self.postponeButton = QPushButton(parent=self.page)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.postponeButton.sizePolicy().hasHeightForWidth())
        self.postponeButton.setSizePolicy(sizePolicy)
        self.postponeButton.setObjectName("postponeButton")
        self.gridLayout_2.addWidget(self.postponeButton, 2, 1, 1, 3)
        self.stackedWidget.addWidget(self.page)
        self.page_2 = QWidget()
        self.page_2.setObjectName("page_2")
        self.verticalLayout = QVBoxLayout(self.page_2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.thanksLabel = QLabel(parent=self.page_2)
        self.thanksLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.thanksLabel.setObjectName("thanksLabel")
        self.verticalLayout.addWidget(self.thanksLabel)
        self.stackedWidget.addWidget(self.page_2)
        self.verticalLayout_2.addWidget(self.stackedWidget)

        self.retranslateUi(DonatePrompt)
        self.stackedWidget.setCurrentIndex(0)
        DonatePrompt.setTabOrder(self.donateButton, self.postponeButton)
        DonatePrompt.setTabOrder(self.postponeButton, self.byeButton)

    def retranslateUi(self, DonatePrompt):
        self.label.setText(_p("DonatePrompt", "Hi! I hope you’re enjoying {app}.\n"
"\n"
"If {app} helps you get work done, please consider making a small donation! This will encourage me to keep developing it. Thank you!\n"
"\n"
"-Iliyas"))
        self.donateButton.setText(_p("DonatePrompt", "Donate {amount}"))
        self.byeButton.setText(_p("DonatePrompt", "Never show this again"))
        self.postponeButton.setText(_p("DonatePrompt", "Remind me in {month}"))
        self.thanksLabel.setText(_p("DonatePrompt", "THANK YOU!"))
