# Form implementation generated from reading ui file 'deletetagdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_DeleteTagDialog(object):
    def setupUi(self, DeleteTagDialog):
        DeleteTagDialog.setObjectName("DeleteTagDialog")
        DeleteTagDialog.setWindowModality(Qt.WindowModality.WindowModal)
        DeleteTagDialog.resize(405, 303)
        self.verticalLayout = QVBoxLayout(DeleteTagDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QLabel(parent=DeleteTagDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pushCheckBox = QCheckBox(parent=DeleteTagDialog)
        self.pushCheckBox.setObjectName("pushCheckBox")
        self.horizontalLayout.addWidget(self.pushCheckBox)
        self.remoteComboBox = QComboBox(parent=DeleteTagDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteComboBox.sizePolicy().hasHeightForWidth())
        self.remoteComboBox.setSizePolicy(sizePolicy)
        self.remoteComboBox.setObjectName("remoteComboBox")
        self.horizontalLayout.addWidget(self.remoteComboBox)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.buttonBox = QDialogButtonBox(parent=DeleteTagDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(DeleteTagDialog)
        self.buttonBox.accepted.connect(DeleteTagDialog.accept)
        self.buttonBox.rejected.connect(DeleteTagDialog.reject)
        self.pushCheckBox.toggled['bool'].connect(self.remoteComboBox.setEnabled)

    def retranslateUi(self, DeleteTagDialog):
        DeleteTagDialog.setWindowTitle(_p("DeleteTagDialog", "Delete tag"))
        self.label.setText(_p("DeleteTagDialog", "Really delete tag {0}?"))
        self.pushCheckBox.setText(_p("DeleteTagDialog", "&Push deletion to:"))
