# Form implementation generated from reading ui file 'conflictview.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_ConflictView(object):
    def setupUi(self, ConflictView):
        ConflictView.setObjectName("ConflictView")
        ConflictView.resize(600, 400)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ConflictView.sizePolicy().hasHeightForWidth())
        ConflictView.setSizePolicy(sizePolicy)
        self.mainLayout = QGridLayout(ConflictView)
        self.mainLayout.setObjectName("mainLayout")
        self.explainer = QLabel(parent=ConflictView)
        self.explainer.setText("<b>BLURB</b><br>BLURB")
        self.explainer.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.explainer.setWordWrap(True)
        self.explainer.setObjectName("explainer")
        self.mainLayout.addWidget(self.explainer, 3, 0, 1, 1)
        spacerItem = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.mainLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.titleLabel = QLabel(parent=ConflictView)
        self.titleLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.titleLabel.setWordWrap(True)
        self.titleLabel.setObjectName("titleLabel")
        self.mainLayout.addWidget(self.titleLabel, 0, 0, 1, 1)
        self.frame = QFrame(parent=ConflictView)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setMaximumSize(QSize(512, 16777215))
        self.frame.setObjectName("frame")
        self.frameLayout = QGridLayout(self.frame)
        self.frameLayout.setContentsMargins(0, 0, 0, 0)
        self.frameLayout.setObjectName("frameLayout")
        self.oursTitleLayout = QHBoxLayout()
        self.oursTitleLayout.setObjectName("oursTitleLayout")
        spacerItem1 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.oursTitleLayout.addItem(spacerItem1)
        self.oursIcon = QLabel(parent=self.frame)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.oursIcon.sizePolicy().hasHeightForWidth())
        self.oursIcon.setSizePolicy(sizePolicy)
        self.oursIcon.setText("ICON")
        self.oursIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursIcon.setObjectName("oursIcon")
        self.oursTitleLayout.addWidget(self.oursIcon)
        self.oursLabel = QLabel(parent=self.frame)
        self.oursLabel.setEnabled(False)
        self.oursLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursLabel.setObjectName("oursLabel")
        self.oursTitleLayout.addWidget(self.oursLabel)
        self.oursPreviewButton = QToolButton(parent=self.frame)
        self.oursPreviewButton.setAutoRaise(True)
        self.oursPreviewButton.setObjectName("oursPreviewButton")
        self.oursTitleLayout.addWidget(self.oursPreviewButton)
        spacerItem2 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.oursTitleLayout.addItem(spacerItem2)
        self.frameLayout.addLayout(self.oursTitleLayout, 0, 0, 1, 1)
        self.theirsButton = QPushButton(parent=self.frame)
        self.theirsButton.setMinimumSize(QSize(0, 48))
        self.theirsButton.setText("THEIRS")
        self.theirsButton.setObjectName("theirsButton")
        self.frameLayout.addWidget(self.theirsButton, 1, 2, 1, 1)
        self.orLabel = QLabel(parent=self.frame)
        self.orLabel.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.orLabel.sizePolicy().hasHeightForWidth())
        self.orLabel.setSizePolicy(sizePolicy)
        self.orLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.orLabel.setObjectName("orLabel")
        self.frameLayout.addWidget(self.orLabel, 1, 1, 1, 1)
        self.oursArrow = QLabel(parent=self.frame)
        self.oursArrow.setEnabled(False)
        self.oursArrow.setText("↘")
        self.oursArrow.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursArrow.setObjectName("oursArrow")
        self.frameLayout.addWidget(self.oursArrow, 2, 0, 1, 1)
        self.theirsTitleLayout = QHBoxLayout()
        self.theirsTitleLayout.setObjectName("theirsTitleLayout")
        spacerItem3 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.theirsTitleLayout.addItem(spacerItem3)
        self.theirsIcon = QLabel(parent=self.frame)
        self.theirsIcon.setText("ICON")
        self.theirsIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsIcon.setObjectName("theirsIcon")
        self.theirsTitleLayout.addWidget(self.theirsIcon)
        self.theirsLabel = QLabel(parent=self.frame)
        self.theirsLabel.setEnabled(False)
        self.theirsLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsLabel.setObjectName("theirsLabel")
        self.theirsTitleLayout.addWidget(self.theirsLabel)
        self.theirsPreviewButton = QToolButton(parent=self.frame)
        self.theirsPreviewButton.setAutoRaise(True)
        self.theirsPreviewButton.setObjectName("theirsPreviewButton")
        self.theirsTitleLayout.addWidget(self.theirsPreviewButton)
        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.theirsTitleLayout.addItem(spacerItem4)
        self.frameLayout.addLayout(self.theirsTitleLayout, 0, 2, 1, 1)
        self.theirsArrow = QLabel(parent=self.frame)
        self.theirsArrow.setEnabled(False)
        self.theirsArrow.setText("↙")
        self.theirsArrow.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsArrow.setObjectName("theirsArrow")
        self.frameLayout.addWidget(self.theirsArrow, 2, 2, 1, 1)
        self.oursButton = QPushButton(parent=self.frame)
        self.oursButton.setMinimumSize(QSize(0, 48))
        self.oursButton.setText("OURS")
        self.oursButton.setObjectName("oursButton")
        self.frameLayout.addWidget(self.oursButton, 1, 0, 1, 1)
        self.stackedWidget = QStackedWidget(parent=self.frame)
        self.stackedWidget.setObjectName("stackedWidget")
        self.mergePage = QWidget()
        self.mergePage.setObjectName("mergePage")
        self.mergeLayout = QGridLayout(self.mergePage)
        self.mergeLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeLayout.setObjectName("mergeLayout")
        self.mergeToolButton = QToolButton(parent=self.mergePage)
        self.mergeToolButton.setAutoRaise(True)
        self.mergeToolButton.setObjectName("mergeToolButton")
        self.mergeLayout.addWidget(self.mergeToolButton, 2, 1, 1, 1)
        self.mergeButton = QPushButton(parent=self.mergePage)
        self.mergeButton.setMinimumSize(QSize(0, 48))
        self.mergeButton.setObjectName("mergeButton")
        self.mergeLayout.addWidget(self.mergeButton, 0, 0, 1, 2)
        self.mergeToolStatus = QLabel(parent=self.mergePage)
        self.mergeToolStatus.setText("MERGE TOOL STATUS")
        self.mergeToolStatus.setWordWrap(True)
        self.mergeToolStatus.setObjectName("mergeToolStatus")
        self.mergeLayout.addWidget(self.mergeToolStatus, 2, 0, 1, 1)
        spacerItem5 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeLayout.addItem(spacerItem5, 4, 1, 1, 1)
        self.stackedWidget.addWidget(self.mergePage)
        self.mergeInProgressPage = QWidget()
        self.mergeInProgressPage.setObjectName("mergeInProgressPage")
        self.mergeInProgressLayout = QGridLayout(self.mergeInProgressPage)
        self.mergeInProgressLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeInProgressLayout.setObjectName("mergeInProgressLayout")
        spacerItem6 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeInProgressLayout.addItem(spacerItem6, 3, 0, 1, 1)
        self.progressBar = QProgressBar(parent=self.mergeInProgressPage)
        self.progressBar.setMaximum(0)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.mergeInProgressLayout.addWidget(self.progressBar, 1, 0, 1, 1)
        self.cancelMergeInProgress = QPushButton(parent=self.mergeInProgressPage)
        self.cancelMergeInProgress.setObjectName("cancelMergeInProgress")
        self.mergeInProgressLayout.addWidget(self.cancelMergeInProgress, 1, 1, 1, 1)
        self.mergeInProgressLabel = QLabel(parent=self.mergeInProgressPage)
        self.mergeInProgressLabel.setText("MERGE IN PROGRESS")
        self.mergeInProgressLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.mergeInProgressLabel.setWordWrap(True)
        self.mergeInProgressLabel.setObjectName("mergeInProgressLabel")
        self.mergeInProgressLayout.addWidget(self.mergeInProgressLabel, 0, 0, 1, 2)
        self.stackedWidget.addWidget(self.mergeInProgressPage)
        self.mergeCompletePage = QWidget()
        self.mergeCompletePage.setObjectName("mergeCompletePage")
        self.mergeCompleteLayout = QGridLayout(self.mergeCompletePage)
        self.mergeCompleteLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeCompleteLayout.setObjectName("mergeCompleteLayout")
        spacerItem7 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeCompleteLayout.addItem(spacerItem7, 6, 1, 1, 1)
        self.discardMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.discardMergeButton.setObjectName("discardMergeButton")
        self.mergeCompleteLayout.addWidget(self.discardMergeButton, 3, 0, 1, 1)
        self.confirmMergeLabel = QLabel(parent=self.mergeCompletePage)
        self.confirmMergeLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.confirmMergeLabel.setWordWrap(True)
        self.confirmMergeLabel.setObjectName("confirmMergeLabel")
        self.mergeCompleteLayout.addWidget(self.confirmMergeLabel, 0, 0, 1, 2)
        self.reworkMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.reworkMergeButton.setObjectName("reworkMergeButton")
        self.mergeCompleteLayout.addWidget(self.reworkMergeButton, 3, 1, 1, 1)
        self.confirmMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.confirmMergeButton.setMinimumSize(QSize(0, 48))
        self.confirmMergeButton.setDefault(True)
        self.confirmMergeButton.setObjectName("confirmMergeButton")
        self.mergeCompleteLayout.addWidget(self.confirmMergeButton, 1, 0, 1, 2)
        self.stackedWidget.addWidget(self.mergeCompletePage)
        self.confirmDeletionPage = QWidget()
        self.confirmDeletionPage.setObjectName("confirmDeletionPage")
        self.confirmDeletionLayout = QGridLayout(self.confirmDeletionPage)
        self.confirmDeletionLayout.setContentsMargins(0, 0, 0, 0)
        self.confirmDeletionLayout.setObjectName("confirmDeletionLayout")
        self.confirmDeletionButton = QPushButton(parent=self.confirmDeletionPage)
        self.confirmDeletionButton.setObjectName("confirmDeletionButton")
        self.confirmDeletionLayout.addWidget(self.confirmDeletionButton, 0, 0, 1, 1)
        spacerItem8 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.confirmDeletionLayout.addItem(spacerItem8, 1, 0, 1, 1)
        self.stackedWidget.addWidget(self.confirmDeletionPage)
        self.emptyPage = QWidget()
        self.emptyPage.setObjectName("emptyPage")
        self.emptyPageLayout = QVBoxLayout(self.emptyPage)
        self.emptyPageLayout.setContentsMargins(0, 0, 0, 0)
        self.emptyPageLayout.setSpacing(0)
        self.emptyPageLayout.setObjectName("emptyPageLayout")
        spacerItem9 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.emptyPageLayout.addItem(spacerItem9)
        self.stackedWidget.addWidget(self.emptyPage)
        self.frameLayout.addWidget(self.stackedWidget, 3, 0, 1, 3)
        self.frameLayout.setColumnStretch(0, 1)
        self.frameLayout.setColumnStretch(2, 1)
        self.mainLayout.addWidget(self.frame, 6, 0, 1, 1)
        spacerItem10 = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        self.mainLayout.addItem(spacerItem10, 5, 0, 1, 1)

        self.retranslateUi(ConflictView)
        self.stackedWidget.setCurrentIndex(0)
        ConflictView.setTabOrder(self.oursButton, self.theirsButton)
        ConflictView.setTabOrder(self.theirsButton, self.mergeButton)
        ConflictView.setTabOrder(self.mergeButton, self.mergeToolButton)
        ConflictView.setTabOrder(self.mergeToolButton, self.cancelMergeInProgress)
        ConflictView.setTabOrder(self.cancelMergeInProgress, self.confirmMergeButton)
        ConflictView.setTabOrder(self.confirmMergeButton, self.discardMergeButton)
        ConflictView.setTabOrder(self.discardMergeButton, self.reworkMergeButton)
        ConflictView.setTabOrder(self.reworkMergeButton, self.confirmDeletionButton)

    def retranslateUi(self, ConflictView):
        ConflictView.setWindowTitle(_p("ConflictView", "Merge conflict"))
        self.titleLabel.setText(_p("ConflictView", "Merge conflict on {0}"))
        self.oursLabel.setText(_p("ConflictView", "OUR version"))
        self.oursPreviewButton.setToolTip(_p("ConflictView", "Preview"))
        self.oursPreviewButton.setText(_p("ConflictView", "Preview"))
        self.orLabel.setText(_p("ConflictView", "or"))
        self.theirsLabel.setText(_p("ConflictView", "THEIR version"))
        self.theirsPreviewButton.setToolTip(_p("ConflictView", "Preview"))
        self.theirsPreviewButton.setText(_p("ConflictView", "Preview"))
        self.mergeToolButton.setToolTip(_p("ConflictView", "Select another external merge tool in the Settings."))
        self.mergeToolButton.setText(_p("ConflictView", "Change merge tool…"))
        self.mergeButton.setToolTip(_p("ConflictView", "<p>Open {tool} so you can merge both versions by hand. When you are done, save the file and quit {tool} to resolve the conflict."))
        self.mergeButton.setText(_p("ConflictView", "Merge both versions in {tool}"))
        self.cancelMergeInProgress.setText(_p("ConflictView", "Abort"))
        self.discardMergeButton.setToolTip(_p("ConflictView", "<p>Ignore the merge that you made in {tool} so that you can start over."))
        self.discardMergeButton.setText(_p("ConflictView", "Discard this merge"))
        self.confirmMergeLabel.setText(_p("ConflictView", "It looks like you’ve finished merging this file."))
        self.reworkMergeButton.setToolTip(_p("ConflictView", "<p>Merge this file in {tool} again, starting where you left off."))
        self.reworkMergeButton.setText(_p("ConflictView", "Merge again"))
        self.confirmMergeButton.setToolTip(_p("ConflictView", "<p>Resolve this conflict by accepting the merge that you’ve completed in {tool}."))
        self.confirmMergeButton.setText(_p("ConflictView", "Resolve the conflict with your merge"))
        self.confirmDeletionButton.setText(_p("ConflictView", "Confirm deletion"))
