# Form implementation generated from reading ui file 'pushdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_PushDialog(object):
    def setupUi(self, PushDialog):
        PushDialog.setObjectName("PushDialog")
        PushDialog.setWindowModality(Qt.WindowModality.NonModal)
        PushDialog.setEnabled(True)
        PushDialog.resize(512, 462)
        PushDialog.setSizeGripEnabled(False)
        PushDialog.setModal(True)
        self.formLayout = QFormLayout(PushDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.localBranchLabel = QLabel(parent=PushDialog)
        self.localBranchLabel.setObjectName("localBranchLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.localBranchLabel)
        self.localBranchEdit = QComboBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.localBranchEdit.sizePolicy().hasHeightForWidth())
        self.localBranchEdit.setSizePolicy(sizePolicy)
        self.localBranchEdit.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.localBranchEdit.setObjectName("localBranchEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.localBranchEdit)
        self.remoteBranchLabel = QLabel(parent=PushDialog)
        self.remoteBranchLabel.setObjectName("remoteBranchLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.remoteBranchLabel)
        self.remoteBranchEdit = QComboBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteBranchEdit.sizePolicy().hasHeightForWidth())
        self.remoteBranchEdit.setSizePolicy(sizePolicy)
        self.remoteBranchEdit.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.remoteBranchEdit.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.remoteBranchEdit.setObjectName("remoteBranchEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.remoteBranchEdit)
        self.newRemoteBranchStackedWidget = QStackedWidget(parent=PushDialog)
        self.newRemoteBranchStackedWidget.setObjectName("newRemoteBranchStackedWidget")
        self.newRemoteBranchPage0 = QWidget()
        self.newRemoteBranchPage0.setObjectName("newRemoteBranchPage0")
        self.horizontalLayout = QHBoxLayout(self.newRemoteBranchPage0)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.remoteNameLabel = QLabel(parent=self.newRemoteBranchPage0)
        self.remoteNameLabel.setText("REMOTE/")
        self.remoteNameLabel.setObjectName("remoteNameLabel")
        self.horizontalLayout.addWidget(self.remoteNameLabel)
        self.newRemoteBranchNameEdit = QLineEdit(parent=self.newRemoteBranchPage0)
        self.newRemoteBranchNameEdit.setText("")
        self.newRemoteBranchNameEdit.setObjectName("newRemoteBranchNameEdit")
        self.horizontalLayout.addWidget(self.newRemoteBranchNameEdit)
        self.newRemoteBranchStackedWidget.addWidget(self.newRemoteBranchPage0)
        self.newRemoteBranchPage1 = QWidget()
        self.newRemoteBranchPage1.setObjectName("newRemoteBranchPage1")
        self.verticalLayout_2 = QVBoxLayout(self.newRemoteBranchPage1)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.forcePushCheckBox = QCheckBox(parent=self.newRemoteBranchPage1)
        self.forcePushCheckBox.setObjectName("forcePushCheckBox")
        self.verticalLayout_2.addWidget(self.forcePushCheckBox)
        self.newRemoteBranchStackedWidget.addWidget(self.newRemoteBranchPage1)
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.newRemoteBranchStackedWidget)
        self.trackCheckBox = QCheckBox(parent=PushDialog)
        self.trackCheckBox.setObjectName("trackCheckBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.trackCheckBox)
        self.trackingLabel = QLabel(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.trackingLabel.sizePolicy().hasHeightForWidth())
        self.trackingLabel.setSizePolicy(sizePolicy)
        self.trackingLabel.setText("<html><head/><body><p><span style=\" font-size:small;\">Extremely long local branch name will track an absurdly long remote branch name instead of another hilariously long remote branch name.</span></p></body></html>")
        self.trackingLabel.setTextFormat(Qt.TextFormat.RichText)
        self.trackingLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.trackingLabel.setWordWrap(True)
        self.trackingLabel.setObjectName("trackingLabel")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.FieldRole, self.trackingLabel)
        spacerItem = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.formLayout.setItem(7, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.groupBox = QGroupBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.statusForm = StatusForm(parent=self.groupBox)
        self.statusForm.setMinimumSize(QSize(0, 32))
        self.statusForm.setObjectName("statusForm")
        self.verticalLayout.addWidget(self.statusForm)
        self.formLayout.setWidget(8, QFormLayout.ItemRole.SpanningRole, self.groupBox)
        self.buttonBox = QDialogButtonBox(parent=PushDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(18, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.localBranchLabel.setBuddy(self.localBranchEdit)
        self.remoteBranchLabel.setBuddy(self.remoteBranchEdit)

        self.retranslateUi(PushDialog)
        self.newRemoteBranchStackedWidget.setCurrentIndex(0)
        self.buttonBox.rejected.connect(PushDialog.reject)
        PushDialog.setTabOrder(self.localBranchEdit, self.remoteBranchEdit)
        PushDialog.setTabOrder(self.remoteBranchEdit, self.forcePushCheckBox)
        PushDialog.setTabOrder(self.forcePushCheckBox, self.newRemoteBranchNameEdit)
        PushDialog.setTabOrder(self.newRemoteBranchNameEdit, self.trackCheckBox)

    def retranslateUi(self, PushDialog):
        PushDialog.setWindowTitle(_p("PushDialog", "Push branch"))
        self.localBranchLabel.setText(_p("PushDialog", "&Local branch:"))
        self.remoteBranchLabel.setText(_p("PushDialog", "Push &to:"))
        self.newRemoteBranchNameEdit.setPlaceholderText(_p("PushDialog", "Branch name on remote"))
        self.forcePushCheckBox.setText(_p("PushDialog", "&Force push"))
        self.trackCheckBox.setText(_p("PushDialog", "&Track this remote branch after pushing"))
from gitfourchette.forms.statusform import StatusForm
