# Form implementation generated from reading ui file 'commitinfodialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CommitInfoDialog(object):
    def setupUi(self, CommitInfoDialog):
        CommitInfoDialog.setObjectName("CommitInfoDialog")
        CommitInfoDialog.resize(480, 342)
        CommitInfoDialog.setWindowTitle("CommitInfoDialog")
        self.gridLayout = QGridLayout(CommitInfoDialog)
        self.gridLayout.setHorizontalSpacing(20)
        self.gridLayout.setObjectName("gridLayout")
        self.iconLabel = QLabel(parent=CommitInfoDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.iconLabel.sizePolicy().hasHeightForWidth())
        self.iconLabel.setSizePolicy(sizePolicy)
        self.iconLabel.setText("ICON")
        self.iconLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.iconLabel.setObjectName("iconLabel")
        self.gridLayout.addWidget(self.iconLabel, 0, 0, 1, 1)
        self.summaryLabel = QLabel(parent=CommitInfoDialog)
        self.summaryLabel.setText("<html><head/><body><p>INFO<br/>INFO<br/>INFO<br/>INFO</p></body></html>")
        self.summaryLabel.setTextFormat(Qt.TextFormat.RichText)
        self.summaryLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.summaryLabel.setWordWrap(True)
        self.summaryLabel.setOpenExternalLinks(False)
        self.summaryLabel.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.summaryLabel.setObjectName("summaryLabel")
        self.gridLayout.addWidget(self.summaryLabel, 0, 1, 2, 1)
        self.buttonBox = QDialogButtonBox(parent=CommitInfoDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 2)
        self.detailsEdit = QTextEdit(parent=CommitInfoDialog)
        self.detailsEdit.setMinimumSize(QSize(0, 200))
        self.detailsEdit.setReadOnly(True)
        self.detailsEdit.setObjectName("detailsEdit")
        self.gridLayout.addWidget(self.detailsEdit, 3, 0, 1, 2)

        self.retranslateUi(CommitInfoDialog)
        self.buttonBox.accepted.connect(CommitInfoDialog.accept)
        self.buttonBox.rejected.connect(CommitInfoDialog.reject)

    def retranslateUi(self, CommitInfoDialog):
        pass
