# Form implementation generated from reading ui file 'checkoutcommitdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CheckoutCommitDialog(object):
    def setupUi(self, CheckoutCommitDialog):
        CheckoutCommitDialog.setObjectName("CheckoutCommitDialog")
        CheckoutCommitDialog.setWindowModality(Qt.WindowModality.NonModal)
        CheckoutCommitDialog.setEnabled(True)
        CheckoutCommitDialog.resize(581, 161)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CheckoutCommitDialog.sizePolicy().hasHeightForWidth())
        CheckoutCommitDialog.setSizePolicy(sizePolicy)
        CheckoutCommitDialog.setSizeGripEnabled(False)
        CheckoutCommitDialog.setModal(True)
        self.verticalLayout = QVBoxLayout(CheckoutCommitDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.mainTitle = QLabel(parent=CheckoutCommitDialog)
        self.mainTitle.setObjectName("mainTitle")
        self.verticalLayout.addWidget(self.mainTitle)
        self.mainGroupBox = QGroupBox(parent=CheckoutCommitDialog)
        self.mainGroupBox.setObjectName("mainGroupBox")
        self.verticalLayout_3 = QVBoxLayout(self.mainGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.switchRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.switchRadioButton.setObjectName("switchRadioButton")
        self.horizontalLayout.addWidget(self.switchRadioButton)
        self.switchComboBox = QComboBox(parent=self.mainGroupBox)
        self.switchComboBox.setObjectName("switchComboBox")
        self.horizontalLayout.addWidget(self.switchComboBox)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.detachHeadRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.detachHeadRadioButton.setObjectName("detachHeadRadioButton")
        self.verticalLayout_3.addWidget(self.detachHeadRadioButton)
        self.resetHeadRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.resetHeadRadioButton.setObjectName("resetHeadRadioButton")
        self.verticalLayout_3.addWidget(self.resetHeadRadioButton)
        self.mergeRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.mergeRadioButton.setObjectName("mergeRadioButton")
        self.verticalLayout_3.addWidget(self.mergeRadioButton)
        self.createBranchRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.createBranchRadioButton.setObjectName("createBranchRadioButton")
        self.verticalLayout_3.addWidget(self.createBranchRadioButton)
        self.verticalLayout.addWidget(self.mainGroupBox)
        self.recurseSubmodulesSpacer = QFrame(parent=CheckoutCommitDialog)
        self.recurseSubmodulesSpacer.setFrameShape(QFrame.Shape.NoFrame)
        self.recurseSubmodulesSpacer.setLineWidth(0)
        self.recurseSubmodulesSpacer.setObjectName("recurseSubmodulesSpacer")
        self.verticalLayout_4 = QVBoxLayout(self.recurseSubmodulesSpacer)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        spacerItem = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)
        self.verticalLayout_4.addItem(spacerItem)
        self.verticalLayout.addWidget(self.recurseSubmodulesSpacer)
        self.recurseSubmodulesTitle = QLabel(parent=CheckoutCommitDialog)
        self.recurseSubmodulesTitle.setObjectName("recurseSubmodulesTitle")
        self.verticalLayout.addWidget(self.recurseSubmodulesTitle)
        self.recurseSubmodulesGroupBox = QGroupBox(parent=CheckoutCommitDialog)
        self.recurseSubmodulesGroupBox.setObjectName("recurseSubmodulesGroupBox")
        self.verticalLayout_2 = QVBoxLayout(self.recurseSubmodulesGroupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.recurseSubmodulesCheckBox = QCheckBox(parent=self.recurseSubmodulesGroupBox)
        self.recurseSubmodulesCheckBox.setChecked(True)
        self.recurseSubmodulesCheckBox.setObjectName("recurseSubmodulesCheckBox")
        self.verticalLayout_2.addWidget(self.recurseSubmodulesCheckBox)
        self.verticalLayout.addWidget(self.recurseSubmodulesGroupBox)
        self.buttonBox = QDialogButtonBox(parent=CheckoutCommitDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(CheckoutCommitDialog)
        self.buttonBox.rejected.connect(CheckoutCommitDialog.reject)
        self.buttonBox.accepted.connect(CheckoutCommitDialog.accept)
        self.switchRadioButton.toggled['bool'].connect(self.switchComboBox.setEnabled)
        CheckoutCommitDialog.setTabOrder(self.switchRadioButton, self.switchComboBox)
        CheckoutCommitDialog.setTabOrder(self.switchComboBox, self.detachHeadRadioButton)
        CheckoutCommitDialog.setTabOrder(self.detachHeadRadioButton, self.resetHeadRadioButton)
        CheckoutCommitDialog.setTabOrder(self.resetHeadRadioButton, self.mergeRadioButton)
        CheckoutCommitDialog.setTabOrder(self.mergeRadioButton, self.createBranchRadioButton)
        CheckoutCommitDialog.setTabOrder(self.createBranchRadioButton, self.recurseSubmodulesCheckBox)

    def retranslateUi(self, CheckoutCommitDialog):
        CheckoutCommitDialog.setWindowTitle(_p("CheckoutCommitDialog", "Check out commit"))
        self.mainTitle.setText(_p("CheckoutCommitDialog", "How do you want to check out this commit?"))
        self.switchRadioButton.setText(_p("CheckoutCommitDialog", "Switch to &branch:"))
        self.switchComboBox.setToolTip(_p("CheckoutCommitDialog", "List of branches that point to this commit."))
        self.detachHeadRadioButton.setText(_p("CheckoutCommitDialog", "Enter &detached HEAD here"))
        self.resetHeadRadioButton.setText(_p("CheckoutCommitDialog", "&Reset HEAD here…"))
        self.mergeRadioButton.setText(_p("CheckoutCommitDialog", "&Merge into HEAD…"))
        self.createBranchRadioButton.setText(_p("CheckoutCommitDialog", "Start &new branch here…"))
        self.recurseSubmodulesTitle.setText(_p("CheckoutCommitDialog", "After the checkout:"))
        self.recurseSubmodulesCheckBox.setText(_p("CheckoutCommitDialog", "Update submodules recursively"))
