# Form implementation generated from reading ui file 'statusform.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_StatusForm(object):
    def setupUi(self, StatusForm):
        StatusForm.setObjectName("StatusForm")
        StatusForm.resize(400, 129)
        StatusForm.setWindowTitle("")
        self.verticalLayout = QVBoxLayout(StatusForm)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.stackedWidget = QStackedWidget(parent=StatusForm)
        self.stackedWidget.setObjectName("stackedWidget")
        self.blurbPage = QWidget()
        self.blurbPage.setObjectName("blurbPage")
        self.verticalLayout_3 = QVBoxLayout(self.blurbPage)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.blurbScrollArea = QScrollArea(parent=self.blurbPage)
        self.blurbScrollArea.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.blurbScrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.blurbScrollArea.setWidgetResizable(True)
        self.blurbScrollArea.setObjectName("blurbScrollArea")
        self.blurbScrollAreaWidgetContents = QWidget()
        self.blurbScrollAreaWidgetContents.setGeometry(QRect(0, 0, 388, 117))
        self.blurbScrollAreaWidgetContents.setObjectName("blurbScrollAreaWidgetContents")
        self.verticalLayout_4 = QVBoxLayout(self.blurbScrollAreaWidgetContents)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.blurbLabel = QLabel(parent=self.blurbScrollAreaWidgetContents)
        self.blurbLabel.setText("BLURB")
        self.blurbLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.blurbLabel.setWordWrap(True)
        self.blurbLabel.setObjectName("blurbLabel")
        self.verticalLayout_4.addWidget(self.blurbLabel)
        self.blurbScrollArea.setWidget(self.blurbScrollAreaWidgetContents)
        self.verticalLayout_3.addWidget(self.blurbScrollArea)
        self.stackedWidget.addWidget(self.blurbPage)
        self.progressPage = QWidget()
        self.progressPage.setObjectName("progressPage")
        self.verticalLayout_2 = QVBoxLayout(self.progressPage)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.titleLabel = QLabel(parent=self.progressPage)
        self.titleLabel.setText("TITLE")
        self.titleLabel.setObjectName("titleLabel")
        self.verticalLayout_2.addWidget(self.titleLabel)
        self.statusLabel = QElidedLabel(parent=self.progressPage)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.statusLabel.sizePolicy().hasHeightForWidth())
        self.statusLabel.setSizePolicy(sizePolicy)
        self.statusLabel.setText("STATUS")
        self.statusLabel.setObjectName("statusLabel")
        self.verticalLayout_2.addWidget(self.statusLabel)
        self.commandLabel = QElidedLabel(parent=self.progressPage)
        self.commandLabel.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.commandLabel.sizePolicy().hasHeightForWidth())
        self.commandLabel.setSizePolicy(sizePolicy)
        self.commandLabel.setText("COMMAND")
        self.commandLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.commandLabel.setObjectName("commandLabel")
        self.verticalLayout_2.addWidget(self.commandLabel)
        spacerItem = QSpacerItem(20, 90, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.progressBar = QProgressBar(parent=self.progressPage)
        self.progressBar.setMaximum(0)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.stackedWidget.addWidget(self.progressPage)
        self.verticalLayout.addWidget(self.stackedWidget)

        self.retranslateUi(StatusForm)
        self.stackedWidget.setCurrentIndex(1)

    def retranslateUi(self, StatusForm):
        pass
from gitfourchette.toolbox.qelidedlabel import QElidedLabel
