# Form implementation generated from reading ui file 'aboutdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.setWindowModality(Qt.WindowModality.ApplicationModal)
        AboutDialog.resize(500, 500)
        self.gridLayout = QGridLayout(AboutDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.iconLabel = QLabel(parent=AboutDialog)
        self.iconLabel.setText("ICON")
        self.iconLabel.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.iconLabel.setObjectName("iconLabel")
        self.gridLayout.addWidget(self.iconLabel, 0, 0, 1, 1)
        self.tabWidget = QTabWidget(parent=AboutDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_2 = QVBoxLayout(self.tab)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollArea = QScrollArea(parent=self.tab)
        self.scrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents_2 = QWidget()
        self.scrollAreaWidgetContents_2.setGeometry(QRect(0, 0, 482, 345))
        self.scrollAreaWidgetContents_2.setObjectName("scrollAreaWidgetContents_2")
        self.horizontalLayout = QHBoxLayout(self.scrollAreaWidgetContents_2)
        self.horizontalLayout.setContentsMargins(32, 16, 32, 16)
        self.horizontalLayout.setSpacing(16)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.aboutBlurb = QLabel(parent=self.scrollAreaWidgetContents_2)
        self.aboutBlurb.setText("ABOUT")
        self.aboutBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.aboutBlurb.setWordWrap(True)
        self.aboutBlurb.setOpenExternalLinks(True)
        self.aboutBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.aboutBlurb.setObjectName("aboutBlurb")
        self.horizontalLayout.addWidget(self.aboutBlurb)
        self.mugshot = QLabel(parent=self.scrollAreaWidgetContents_2)
        self.mugshot.setText("MUG")
        self.mugshot.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTop|Qt.AlignmentFlag.AlignTrailing)
        self.mugshot.setObjectName("mugshot")
        self.horizontalLayout.addWidget(self.mugshot)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents_2)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.tabWidget.addTab(self.tab, "")
        self.tab_4 = QWidget()
        self.tab_4.setObjectName("tab_4")
        self.verticalLayout_7 = QVBoxLayout(self.tab_4)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.scrollArea_4 = QScrollArea(parent=self.tab_4)
        self.scrollArea_4.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea_4.setWidgetResizable(True)
        self.scrollArea_4.setObjectName("scrollArea_4")
        self.scrollAreaWidgetContents_4 = QWidget()
        self.scrollAreaWidgetContents_4.setGeometry(QRect(0, 0, 117, 48))
        self.scrollAreaWidgetContents_4.setObjectName("scrollAreaWidgetContents_4")
        self.verticalLayout_6 = QVBoxLayout(self.scrollAreaWidgetContents_4)
        self.verticalLayout_6.setContentsMargins(32, 16, 32, 16)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.ackBlurb = QLabel(parent=self.scrollAreaWidgetContents_4)
        self.ackBlurb.setText("LICENSE")
        self.ackBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.ackBlurb.setWordWrap(True)
        self.ackBlurb.setOpenExternalLinks(True)
        self.ackBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard|Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextBrowserInteraction|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.ackBlurb.setObjectName("ackBlurb")
        self.verticalLayout_6.addWidget(self.ackBlurb)
        self.scrollArea_4.setWidget(self.scrollAreaWidgetContents_4)
        self.verticalLayout_7.addWidget(self.scrollArea_4)
        self.tabWidget.addTab(self.tab_4, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName("tab_3")
        self.verticalLayout = QVBoxLayout(self.tab_3)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea_3 = QScrollArea(parent=self.tab_3)
        self.scrollArea_3.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea_3.setWidgetResizable(True)
        self.scrollArea_3.setObjectName("scrollArea_3")
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 117, 48))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_5 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_5.setContentsMargins(32, 16, 32, 16)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.licenseBlurb = QLabel(parent=self.scrollAreaWidgetContents)
        self.licenseBlurb.setText("LICENSE")
        self.licenseBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.licenseBlurb.setWordWrap(True)
        self.licenseBlurb.setOpenExternalLinks(True)
        self.licenseBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard|Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextBrowserInteraction|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.licenseBlurb.setObjectName("licenseBlurb")
        self.verticalLayout_5.addWidget(self.licenseBlurb)
        self.scrollArea_3.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea_3)
        self.tabWidget.addTab(self.tab_3, "")
        self.gridLayout.addWidget(self.tabWidget, 3, 0, 1, 4)
        self.buttonBox = QDialogButtonBox(parent=AboutDialog)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 4, 3, 1, 1)
        self.header = QLabel(parent=AboutDialog)
        self.header.setText("HEADER")
        self.header.setOpenExternalLinks(True)
        self.header.setObjectName("header")
        self.gridLayout.addWidget(self.header, 0, 2, 1, 2)
        spacerItem = QSpacerItem(8, 20, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 1, 1, 1)
        spacerItem1 = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.gridLayout.addItem(spacerItem1, 2, 2, 1, 1)
        self.versionLabel = QLabel(parent=AboutDialog)
        self.versionLabel.setText("VERSION")
        self.versionLabel.setObjectName("versionLabel")
        self.gridLayout.addWidget(self.versionLabel, 1, 2, 1, 2)

        self.retranslateUi(AboutDialog)
        self.tabWidget.setCurrentIndex(0)
        self.buttonBox.accepted.connect(AboutDialog.close)

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(_p("AboutDialog", "About {0}"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _p("AboutDialog", "About"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), _p("AboutDialog", "Acknowledgments"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), _p("AboutDialog", "License"))
